--[[
    Core
]]

local ADDON_NAME, ns = ...

local Printer = ns.Printer
local displays = ns.displays
local display = ns.display
local settings = ns.settings
local L = ns.L

local LDB = LibStub("LibDataBroker-1.1")
local dataobj = LDB:GetDataObjectByName("OakInterrupts")

--------------------------------------------------------------------------------

local function forall(func)
    for _, v in pairs(displays) do
        func(v)
    end
end

local addon = CreateFrame("Frame")

function addon:Initialise()
    Printer.debug(1, "Debug level: %d", Printer.debug_level)
    OIDDB = OIDDB or {}
    if CharDisplays then
        for k, v in pairs(CharDisplays) do
            ns.CharDisplays[k] = v
        end
        display.RestoreDisplays(ns.CharDisplays)
    end
    self:UnregisterEvent("ADDON_LOADED")
end

function addon:ADDON_LOADED(addon_name)
    if addon_name == ADDON_NAME then
        self:Initialise()
    end
end

function addon:PLAYER_LOGOUT(self)
    CharDisplays = ns.CharDisplays
    -- Save display positions
    for name, opts in pairs(CharDisplays) do
        local point, rel_to, rel_point, xoff, yoff = displays[name]:GetPoint()
        rel_to = rel_to and rel_to:GetName()
        CharDisplays[name].position = { point, rel_to, rel_point, xoff, yoff }
    end
end

function addon:PLAYER_REGEN_DISABLED(self)
    forall(function(disp)
               display.Lock(disp)
               if #disp.queue.frames > 0 and disp.rotation then
                   disp.rotation.next_index = 1
                   ns.rotation.AdvanceRotation(disp, dataobj)
                   local fs = disp.queue.frames
                   display.ColourInterrupter(fs[1], settings.colour.highlight)
               end
           end)
end

function addon:PLAYER_REGEN_ENABLED(self)
    forall(function(disp)
               display.Unlock(disp)
               if #disp.queue.frames > 0 and disp.rotation then
                   local n = disp.rotation.next_index
                   local fs = disp.queue.frames
                   display.ColourInterrupter(fs[n], settings.colour.default)
               end
           end)
end

function addon:GROUP_ROSTER_UPDATE(self)
    if IsInGroup() then
        forall(function(disp)
                    if not disp.inactive and not disp.hidden then
                        display.Show(disp, true)
                    end
                end)
    else
        forall(function(disp)
                    if disp.temporary then
                        display.Close(disp)
                    else
                        display.Hide(disp, true)
                    end
                end)
    end
end

function addon:PLAYER_LOGIN(self)
    forall(function(disp)
            if not IsInGroup() then
                display.Hide(disp, true)
            end
           end)
end

function addon:CHAT_MSG_ADDON(self, ...)
    ns.Comms.HandleMessage(...)
end

addon:RegisterEvent("ADDON_LOADED")
addon:RegisterEvent("PLAYER_REGEN_DISABLED")
addon:RegisterEvent("PLAYER_REGEN_ENABLED")
addon:RegisterEvent("PLAYER_LOGOUT")
addon:RegisterEvent("GROUP_ROSTER_UPDATE")
addon:RegisterEvent("PLAYER_LOGIN")
addon:RegisterEvent("CHAT_MSG_ADDON")
addon:SetScript("OnEvent", function(self, event, ...) self[event](self, ...) end)

ns.OID = addon

--- ConfigMode ---
CONFIGMODE_CALLBACKS = CONFIGMODE_CALLBACKS or {}
local ConfigModeHandler
do
    local wasHidden = {}
    local wasInGroup
    ConfigModeHandler = function(action)
        if action == "ON" then
            wasInGroup = IsInGroup()
            forall(function(disp)
                wasHidden = not disp:IsShown()
                display.Show(disp, true)
            end)
        elseif action == "OFF" then
            forall(function(disp)
                -- only hide if all frames should be hidden (solo) or we
                -- explicitly hid a frame that normally shows in a group.
                -- wasInGroup makes sure we don't hide frames if we joined a
                -- group between calling 'ON' and 'OFF'
                if not IsInGroup() or (wasHidden[disp] and wasInGroup) then
                    display.Hide(disp, true)
                end
            end)
        end
    end
end
CONFIGMODE_CALLBACKS[ADDON_NAME] = ConfigModeHandler

--------------------------------------------------------------------------------
-- User Control
--------------------------------------------------------------------------------

-- Keybindings
BINDING_NAME_OID_RESET_HOTKEY = L["Reset rotations"]
BINDING_HEADER_OID = ns.title

function OIDResetRotations()
    forall(
        function(disp)
            if disp.hotkey_reset then
                ns.rotation.Reset(disp)
                ns.Comms.SendReset(disp)
            end
        end)
end

-- Slash commands
local commands = LibStub:GetLibrary("OakSlash-1.0")(ns.title)
local t = {
    new = display.New,
    delete = display.Close,
    show = display.Show,
    hide = display.Hide,
    toggle = display.ToggleVisibility,
    autofill = function(disp) display.SetAutoInclude(disp, true) end,
    noautofill = function(disp) display.SetAutoInclude(disp, false) end,
    rename = display.Rename,
    refill = display.Refill,
    rotation = function(disp) display.SetRotation(disp, true) end,
    norotation = function(disp) display.SetRotation(disp, false) end,
    announce = function(disp) display.SetRotation(disp, "ANNOUNCE") end,
    share = display.ShareRotation,
    debug = function(l) Printer.SetDebug(l); ns.rotation.Printer.SetDebug(l);
                        ns.FramePool.Printer.SetDebug(l) end,
    grow = { up = function(disp) display.SetDirection(disp, "up") end,
             down = function(disp) display.SetDirection(disp, "down") end,
         },
    lock = display.Lock,
    unlock = display.Unlock,
    bind = function(disp) display.SetHotkeyReset(disp, true) end,
    nobind = function(disp) display.SetHotkeyReset(disp, nil) end,
    config = CONFIGMODE_CALLBACKS[ADDON_NAME],
    list = display.PrintNames,
    minimap = { show = function() addon.Minimap:Hide(false) end,
                hide = function() addon.Minimap:Hide(true) end }
}
commands:NewCommand("oi", t)
--------------------------------------------------------------------------------
