
local ADDON_NAME, ns = ...
local rotation = ns.rotation
local rearrange = ns.rearrange
local FramePool = ns.FramePool
local Menu = ns.Menu
local settings = ns.settings
local lib = ns.lib

local L = ns.L
local title = select(2, GetAddOnInfo(ADDON_NAME))
local Printer = LibStub:GetLibrary("OakPrint-1.0")(title)

local display = {}
local CharDisplays = {}
local opts = settings.display or {}
--------------------------------------------------------------------------------
local GetTime = GetTime

--------------------------------------------------------------------------------
-- Libraries
--------------------------------------------------------------------------------

-- datafeed
local LDB = LibStub("LibDataBroker-1.1")
local dataobj = LDB:GetDataObjectByName("OakInterrupts")

local function indexOf(t, val)
    for i, v in ipairs(t) do
        if v == val then
            return i
        end
    end
end

--------------------------------------------------------------------------------
-- Lookup tables
--------------------------------------------------------------------------------

local class = setmetatable({}, { __index = function(t, k)
        local cls = select(2, UnitClass(k))
        t[k] = cls
        return cls
    end})

local displayof = {}
local DispByName = {}

--------------------------------------------------------------------------------
-- Interrupter display functions
--------------------------------------------------------------------------------

local default_colour = ns.settings.colour.default
local highlight_colour = ns.settings.colour.highlight
local disabled_colour = ns.settings.colour.disabled

function display.ColourInterrupter(frame, colour, alpha)
    frame.bg:SetColorTexture(unpack(colour))
    frame:SetAlpha(alpha or 1)
end

function UnitDisabled(unit)
    return UnitIsDead(unit) or not UnitIsConnected(unit)
end

function display.RotationHighlight(disp)
    if not disp.rotation then return end
    for i, frame in ipairs(disp.queue.frames) do
        if i == disp.rotation.next_index and UnitAffectingCombat('player') then
            display.ColourInterrupter(frame, highlight_colour)
        elseif not UnitDisabled(frame.name) then
            display.ColourInterrupter(frame, default_colour)
        end
    end
end

local function SetNameText(frame, name, class)
    local colour = "|c" .. lib.classHexColour(class)
    frame.txtname:SetText(colour .. lib.formatName(name) .. "|r")
end

local function SetTimeText(frame, time)
    local colour
    local t = type(time) == "number" and lib.formatTime(time) or time
    if type(time) == "number" and time > 0 then
        colour = "|cFFFF0000"
    else
        if t == "READY" then
            colour = "|cFF00FF00"
        else
            colour = "|cFF666666" -- OFFLINE or DEAD
        end
        t = L[t]
    end
    frame.time:SetText(colour .. t .. "|r")
end

local function SetInterrupter(frame, name)
    frame.name = name
    local class = class[name]
    SetNameText(frame, name, class)
end

local function OnMouseDown(frame, btn)
    local disp = displayof[frame]
    if btn == "RightButton" then
        disp:RemoveInterrupter(frame)
        display.RotationHighlight(disp)
        disp:UpdateLayout()
    elseif btn == "LeftButton" then
        rearrange.BeginMoving(disp.queue, frame)
    end
end

local function OnMouseUp(frame, btn)
    if btn == "LeftButton" and frame.isMoving then
        rearrange.EndMoving(displayof[frame].queue, frame)
    end
end


local function OnUpdate(frame, elapsed)
    local disp = displayof[frame]
    local new_d = rearrange.MouseUpdate(disp, frame, elapsed)
    if new_d then
        displayof[frame] = new_d
        disp:ResizeDisplay()
        new_d:ResizeDisplay()
    end
    -- update time display
    if dataobj[frame.name] then
        if not UnitIsConnected(frame.name) then
            display.ColourInterrupter(frame, disabled_colour, 0.4)
            SetTimeText(frame, "OFFLINE")
        elseif UnitIsDead(frame.name) then
            display.ColourInterrupter(frame, disabled_colour, 0.4)
            SetTimeText(frame, "DEAD")
        else
            display.ColourInterrupter(frame, default_colour)
            SetTimeText(frame, dataobj[frame.name] - GetTime())
        end
    end
    display.RotationHighlight(disp)
end

local function HeadingMouseDown(frame, btn)
    local disp = frame:GetParent()
    if btn == "LeftButton" then
        disp.premoveref = { disp:GetPoint() }
        disp:StartMoving()
        disp:SetUserPlaced(false)
        _, _, _, disp.x, disp.y = disp:GetPoint()
    end
end

local function HeadingMouseUp(frame, btn)
    local disp = frame:GetParent()
    if btn == "LeftButton" then
        local _, _, _, x, y = disp:GetPoint()
        local es = disp:GetEffectiveScale()
        x = x - disp.x
        y = y - disp.y
        disp:StopMovingOrSizing()
        disp:ClearAllPoints()
        local p, rf, rp, xoff, yoff = unpack(disp.premoveref)
        disp:SetPoint(p, rf, rp, xoff + x, yoff + y)
        disp.x, disp.y, disp.premoveref = nil, nil, nil
    end
end


--------------------------------------------------------------------------------
-- FramePools
--------------------------------------------------------------------------------

local display_pool = FramePool:NewPool("OakInterrupts_Display", UIParent)
local interrupter_pool = FramePool:NewPool("OakInterrupts_InterrupterFrame", UIParent)

--------------------------------------------------------------------------------
-- Create interrupter frames
--------------------------------------------------------------------------------

local function LayoutFrame(disp, frame)
    frame:SetWidth(disp.width - 2 * disp.padding)
    frame:SetHeight(disp.player_height)
end

local function InitFrame(disp, frame, name)
    local bg = frame:CreateTexture()
    bg:SetAllPoints()
    frame.bg = bg

    local txtname = frame:CreateFontString()
    txtname:SetPoint("LEFT", frame, "LEFT", 5, 0)
    txtname:SetFontObject("GameFontNormal")
    frame.txtname = txtname

    local time = frame:CreateFontString()
    time:SetPoint("RIGHT", frame, "RIGHT", -5, 0)
    time:SetFontObject("GameFontNormal")
    frame.time = time

    frame:EnableMouse(true)
    frame:SetMovable(true)
    frame:SetScript("OnUpdate", OnUpdate)
    frame:SetScript("OnMouseDown", OnMouseDown)
    frame:SetScript("OnMouseUp", OnMouseUp)
end

local function CreateInterrupterFrame(disp, name)
    -- call disp:UpdateLayout() after this returns if you want to update
    -- the display.
    local frame, new = interrupter_pool:NewFrame()
    if new then
        InitFrame(disp, frame, name)
    end
    display.ColourInterrupter(frame, default_colour)
    frame:SetParent(disp.container)
    LayoutFrame(disp, frame)
    frame:Show()
    SetInterrupter(frame, name)
    return frame
end


--------------------------------------------------------------------------------
-- Base frames for a display
--------------------------------------------------------------------------------

-- methods --
local function AddInterrupter(disp, name)
    -- call disp:UpdateLayout() after this returns if you want to update
    -- the display.
    if name == "type" or name == "polling" then return end
    assert(not disp.watch_list[name], string.format("Error: %s is already watching %s", disp.name, name))
    Printer.debug(1, L["%s is now watching %s"], disp.name, name)
    local frame = CreateInterrupterFrame(disp, name)
    disp.watch_list[name] = frame
    disp.queue:AddFrame(frame)
    displayof[frame] = disp
    frame:Show()
end

local function RemoveInterrupter(disp, nameorframe)
    assert(type(nameorframe) == "string" or type(nameorframe) == "table",
                "Usage: RemoveInterrupter(display, name | frame)")
    local frame
    local name
    if type(nameorframe) == "string" then
        for _, f in ipairs(disp.queue.frames) do
            if f.name == nameorframe then
                frame = f
                name = nameorframe
            end
        end
    elseif type(nameorframe) == "table" then
        frame = nameorframe
        name = frame.name
    end
    Printer.debug(3, L["removing interrupter - frame: %s, name: %s"], frame:GetName() or tostring(frame), name)

    disp.watch_list[name] = false
    disp.queue:RemoveFrame(frame)
    displayof[frame] = nil
    Printer.debug(1, L["%s no longer watching %s"], disp.name, name)
    interrupter_pool:Recycle(frame)
end

local function DataChanged(disp, eventname, doname, attr, value, dobj)
    if attr == "polling" then return end
    if not dataobj[attr] and disp.watch_list[attr] then
        disp:RemoveInterrupter(attr)
        disp.watch_list[attr] = nil
        disp:UpdateLayout()
    elseif disp.watch_list[attr] == false then
        return
    elseif disp.watch_list[attr] and value > GetTime() then
        Printer.debug(3, L["Interrupt used by %s"], attr)
        rotation.AdvanceRotation(disp, dataobj)
    else -- we are not explicitly ignoring or watching
        if disp.autoinclude then
            disp:AddInterrupter(attr)
            disp:UpdateLayout()
        end
    end
end

local function ChangeName(disp, name)
    if DispByName[name] then
        Printer.print(L["You already have an interrupt list named %s"], name)
        return
    end
    disp.heading.text:SetText(name)

    local prev_name = disp.name
    DispByName[prev_name] = nil
    if not disp.temporary then
        CharDisplays[name] = CharDisplays[prev_name]
        CharDisplays[prev_name] = nil
    end

    disp.name = name
    DispByName[name] = disp
end

local function Refill(disp)
    for k, v in LDB:pairs(dataobj) do
        if not disp.watch_list[k] then
            disp:AddInterrupter(k)
        end
    end
    disp:UpdateLayout()
end

-- layout functions --
local function ResizeDisplay(disp)
    disp:SetWidth(disp.width)
    local pad = (disp.container:GetHeight() == 0 and 2 or 3) * disp.padding
    disp:SetHeight(disp.heading:GetHeight() + disp.container:GetHeight() + pad)
end

local function UpdateLayout(disp)
    disp.queue:LayoutQueue()
    disp:ResizeDisplay(disp)
end

local function LayoutDisplay(disp)
    --only supporting TOP and BOTTOM at the moment
    opposite = { TOP = "BOTTOM", BOTTOM = "TOP" }
    local anchor = disp.queue.anchor
    local sign = anchor == "TOP" and 1 or -1
    local padding = sign * disp.padding

    local point, relframe, relpoint, xoff, yoff = disp:GetPoint()
    if anchor == opposite[point] then
        yoff = yoff + 2 * padding + sign * disp.heading:GetHeight()
    end

    disp:ClearAllPoints()
    disp.heading:ClearAllPoints()
    disp.container:ClearAllPoints()

    disp:UpdateLayout()

    -- Position display
    disp:SetPoint(anchor, relframe, relpoint, xoff, yoff)

    -- layout components
    disp.heading:SetPoint(anchor, disp, anchor, 0, -padding)
    disp.heading:SetPoint("LEFT", disp, "LEFT")
    disp.heading:SetPoint("RIGHT", disp, "RIGHT")

    disp.container:SetPoint(anchor, disp.heading, opposite[anchor], 0, -padding)
end

-- creating a display --
local function CreateDisplay(name, height, width)
    local padding = opts.padding
    local height = height or opts.player_height
    local width = width or opts.player_width

    local disp, new = display_pool:NewFrame()
    do
        local bg = disp:CreateTexture()
        bg:SetAllPoints()
        bg:SetColorTexture(.1,.1,.1,.7)
    end

    if new then
        --- Components ---
        ------------------

        local heading, container

        --- heading bar ---
        heading = CreateFrame("Frame", "$parent_heading", disp)
        disp.heading = heading

        heading:EnableMouse(true)
        heading:SetScript("OnMouseDown", HeadingMouseDown)
        heading:SetScript("OnMouseUp", HeadingMouseUp)

        disp:SetMovable(true)

        -- title
        local text = heading:CreateFontString()
        text:SetPoint("CENTER")
        text:SetFontObject("GameFontNormal")
        text:SetText("j")
        heading.text = text

        --- container ---
        container = CreateFrame("Frame", "$parent_container", disp)
        disp.container = container

        -- Initialise display --
        ------------------------

        --- layout components ---
        do
            -- text
            local hh = heading.text:GetStringHeight()
            heading:SetHeight(hh)

        end

        -- drop down menu
        Menu.Register(disp)

        disp.queue = ns.FrameQueue(container, true)
    end

    disp.heading.text:SetText(name or L["Interrupters"])

    -- layout/display options

    disp.padding = padding
    disp.player_height = height
    disp.width = width
    disp.name = name
    disp.autoinclude = false
    disp.locked = false
    disp.hotkey_reset = false
    disp.watch_list = {} -- the subset of dataobj this display watches

    -- methods
    disp.AddInterrupter = AddInterrupter
    disp.RemoveInterrupter = RemoveInterrupter
    disp.DataChanged = DataChanged
    disp.ChangeName = ChangeName
    disp.Refill = Refill
    disp.ResizeDisplay = ResizeDisplay
    disp.LayoutDisplay = LayoutDisplay
    disp.UpdateLayout = UpdateLayout

    -- position display
    disp:SetPoint("TOP", UIParent, "TOP", 0, -300)
    LayoutDisplay(disp)

    return disp
end

--------------------------------------------------------------------------------
-- Display interface
--------------------------------------------------------------------------------

--- helper functions ---
local function ProcessDisp(disporname)
    -- helper function allowing for so we can provide a name or display
    if type(disporname) == "string" then
        return DispByName[disporname]
    else
        return disporname
    end
end

local function GenerateNewName()
    local count = 1
    while DispByName[L["Interrupters"] .. count] do
        count = count + 1
    end
    return L["Interrupters"] .. count
end
---

function display.New(name, temp)
    if name and DispByName[name] then
        Printer.print(L["You already have an interrupt list named %s"], name)
        return
    end
    local name = name or GenerateNewName()
    local disp = CreateDisplay(name)
    if not temp then
        for k, _ in LDB:pairs(dataobj) do
            disp:AddInterrupter(k)
        end
    end
    disp:Show()
    disp:UpdateLayout()
    LDB.RegisterCallback(disp, "LibDataBroker_AttributeChanged_OakInterrupts", "DataChanged")
    DispByName[name] = disp

    -- set CharDisplays value
    if not temp then
        CharDisplays[name] = {}
        disp.temporary = nil
    else
        disp.temporary = true
    end
end

function display.Close(disp)
    disp = ProcessDisp(disp)
    if not disp then return end
    Printer.debug(1, L["Closing display %s"], disp.name)
    LDB.UnregisterCallback(disp, "LibDataBroker_AttributeChanged_OakInterrupts")
    for k, v in ipairs(disp.queue.frames) do
        RemoveInterrupter(disp, v)
    end
    DispByName[disp.name] = nil
    if not disp.temporary then
        CharDisplays[disp.name] = nil
    else
        ns.Comms.Cleanup(disp)
    end
    display_pool:Recycle(disp)
end

function display.Show(disp, notSave)
    disp = ProcessDisp(disp)
    disp:Show()
    disp.hidden = false
    if not notSave then
        disp.inactive = false
        if not disp.temporary then
            CharDisplays[disp.name].inactive = false
        end
    end
end

function display.Hide(disp, notSave)
    disp = ProcessDisp(disp)
    disp:Hide()
    disp.hidden = true
    if not notSave then
        disp.inactive = true
        if not disp.temporary then
            CharDisplays[disp.name].inactive = true
        end
    end
end

do
    local function getVisible()
        for _, disp in pairs(DispByName) do
            if disp:IsShown() and not disp.inactive then
                return true
            end
        end
        return false
    end

    function display.ToggleVisibility()
        local visible = getVisible()
        for _, disp in pairs(DispByName) do
            if not disp.inactive then
                if visible then
                    display.Hide(disp, true)
                else
                    display.Show(disp, true)
                end
            end
        end
    end
end

function display.Lock(disp, mode)
    disp = ProcessDisp(disp)
    if not disp then return end
    Printer.debug(2, L["Locking display %s"], disp.name)
    disp.heading:EnableMouse(false)
    if mode and string.upper(mode) == "COMBAT" then
        for _, f in ipairs(disp.queue.frames) do
            f:EnableMouse(false)
        end
    else
        disp.locked = true
        if not disp.temporary then
            CharDisplays[disp.name].locked = true
        end
    end
end

function display.Unlock(disp, mode)
    disp = ProcessDisp(disp)
    if not disp then return end
    Printer.debug(2, L["Unlocking display %s"], disp.name)
    if mode and string.upper(mode) == "COMBAT" then
        for _, f in ipairs(disp.queue.frames) do
            f:EnableMouse(true)
        end
        if not disp.locked then
            disp.heading:EnableMouse(true)
        end
    else
        disp.locked = false
        disp.heading:EnableMouse(true)
        if not disp.temporary then
            CharDisplays[disp.name].locked = nil
        end
    end
end

function display.Refill(disp)
    disp = ProcessDisp(disp)
    Printer.debug(1, L["Refilling display %s"], disp.name)
    disp:Refill()
end

function display.ShareRotation(disp)
    disp = ProcessDisp(disp)
    if disp.rotation then
        rotation.AnnounceOrder(disp)
    end
end

function display.Rename(disp, name)
    assert(type(name) == "string", "Usage: Rename(display, name) - must provide a string name")
    disp = ProcessDisp(disp)
    disp:ChangeName(name)
end

--- Display attributes ---
function display.SetDirection(disp, dir)
    disp = ProcessDisp(disp)
    if not disp then return end
    local anchor = dir == "up" and "BOTTOM" or "TOP"
    if anchor == "TOP" then
        dir = "down"
    end

    disp.queue.anchor = anchor
    disp:LayoutDisplay()

    if not disp.temporary then
        CharDisplays[disp.name].grow = dir
    end
end

function display.SetAutoInclude(disp, b)
    assert(type(b) == "string" or type(b) == "boolean",
        "Usage: SetAutoInclude(disp, boolean), the boolean can be given as a string.")
    disp = ProcessDisp(disp)
    if type(b) == "string" then
        if string.upper(b) == "FALSE" then
            b = false
        else
            b = true
        end
    end
    disp.autoinclude = b
    if not disp.temporary then
        CharDisplays[disp.name].autoinclude = b
    end
end

function display.SetRotation(disp, flag)
    disp = ProcessDisp(disp)
    local announce = flag == "ANNOUNCE" and true or false
    Printer.debug(2, L["Enabling rotation for %s: %s"], disp.name, tostring(announce))
    if flag then
        disp.rotation = { next_index = 1, announce = announce }
    else
        disp.rotation = nil
    end
    if not disp.temporary then
        CharDisplays[disp.name].rotation = flag
    end
end

function display.SetHotkeyReset(disp, flag)
    disp = ProcessDisp(disp)
    if type(flag) == "string" then
        flag = string.upper(flag) == "FALSE" and false or true
    end
    disp.hotkey_reset = flag
    if not disp.temporary then
        CharDisplays[disp.name].hotkey_reset = flag
    end
end

local AttrSetter = {
    grow = display.SetDirection,
    autoinclude = display.SetAutoInclude,
    rotation = display.SetRotation,
    locked = function(disp, locked) if locked then display.Lock(disp) end end,
    position = function(disp, t)
        disp:ClearAllPoints()
        disp:SetPoint(unpack(t))
    end,
    hotkey_reset = display.SetHotkeyReset,
    inactive = function(disp, inactive) if inactive then display.Hide(disp) else display.Show(disp) end end
}

local function SetAttribute(disp, attr, val)
    local func = AttrSetter[attr]
    if func then
        func(disp, val)
    end
end

function display.RestoreDisplays(display_table)
    assert(type(display_table) == 'table', "Error: argument is not a table")
    for name, disp in pairs(DispByName) do
        if not display_table[name] then
            display.Close(disp)
        end
    end
    for name, disp_opts in pairs(display_table) do
        local disp = DispByName[name]
        if not disp then
            display.New(name)
            disp = DispByName[name]
        end
        for attr, val in pairs(disp_opts) do
            SetAttribute(disp, attr, val)
        end
    end
end

function display.PrintNames()
    local msg = '\n' .. L["Interrupt displays:"]
    for name, _ in pairs(DispByName) do
        msg = msg .. '\n    - ' .. name
    end
    Printer.print(msg)
end
--------------------------------------------------------------------------------
ns.displays = DispByName
ns.display = display
ns.Printer = Printer
ns.CharDisplays = CharDisplays
ns.title = title

-- TODO:
