local addon_name, ns = ...

local FrameQueue = {}

--------------------------------------------------------------------------------
-- Lookup tables
--------------------------------------------------------------------------------

local opposite = {
    TOP    = "BOTTOM",
    LEFT   = "RIGHT",
    BOTTOM = "TOP",
    RIGHT  = "LEFT"
}

local anchor_point = setmetatable({
    TOP    = "TOP",
    BOTTOM = "BOTTOM",
    LEFT   = "TOPRIGHT",
    RIGHT  = "TOPLEFT"},
    { __index = function(t, k) return k end }
)

local rel_anchor_point = setmetatable({
    TOP    = "BOTTOM",
    BOTTOM = "TOP",
    LEFT   = "BOTTOMRIGHT",
    RIGHT  = "BOTTOMLEFT"},
    { __index = function(t, k) return k end }
)

local coord = {
    TOP    = "y",
    LEFT   = "x",
    BOTTOM = "y",
    RIGHT  = "x"
}

--------------------------------------------------------------------------------
-- FrameQueue
--------------------------------------------------------------------------------

local function tremovebyval(t, o)
    for k, v in pairs(t) do
        if v == o then
            return table.remove(t, k)
        end
    end
end

local function GetPadding(anchor, pad)
    if anchor == "TOP" then return 0, -pad
    elseif anchor == "BOTTOM" then return 0, pad
    elseif anchor == "LEFT" then return pad, 0
    elseif anchor == "RIGHT" then return -pad, 0
    else return 0, 0
    end
end

function FrameQueue:AddFrame(frame)
    assert(not tContains(self.frames, frame), "Error the FrameQueue " .. tostring(self)
            .. " already contains the frame " .. (frame:GetName() or tostring(frame)))
    table.insert(self.frames, frame)
end

function FrameQueue:RemoveFrame(frame)
    tremovebyval(self.frames, frame)
end

function FrameQueue:MoveFrame(frame, i)
    tremovebyval(self.frames, frame)
    table.insert(self.frames, i, frame)
end

function FrameQueue:AnchorFrames()
    local point, rel_frame, rel_point, xoff, yoff = self.container:GetPoint()
    self.container:ClearAllPoints()
    if self.outside and rel_frame ~= UIParent then
        point = anchor_point[self.anchor]
        rel_point = rel_anchor_point[self.anchor]
    end
    self.container:SetPoint(point, rel_frame, rel_point, xoff, yoff)
    local xpad, ypad = GetPadding(self.anchor, self.padding)
    for k, v in pairs(self.frames) do
        v:ClearAllPoints()
        if k > 1 then
            v:SetPoint(self.anchor, self.frames[k-1], opposite[self.anchor], xpad, ypad)
        else
            v:SetPoint(self.anchor, self.container, self.anchor, xpad, ypad)
        end
        if self.unstick then
            v:StartMoving()
            v:StopMovingOrSizing()
            v:SetUserPlaced(false)
        end
    end
end

function FrameQueue:ResizeContainer()
    local numFrames = #self.frames
    local pad = self.padding
    local height = numFrames > 0 and pad or 0
    local width = height
    for _, frame in ipairs(self.frames) do
        if coord[self.anchor] == "y" then
            height = height + frame:GetHeight() + pad
            width = max(width, frame:GetWidth() + 2 * pad)
        elseif coord[self.anchor] == "x" then
            height = max(height, frame:GetHeight() + 2 * pad)
            width = width + frame:GetWidth() + pad
        end
    end

    self.container:SetHeight(height)
    self.container:SetWidth(width)
end

function FrameQueue:LayoutQueue()
    self:ResizeContainer()
    self:AnchorFrames()
end

function FrameQueue:NewFrameQueue(container, outside)
    local queue = { frames = {} }
    if container then
        queue.container = container
    else
        queue.container = CreateFrame("Frame", name, UIParent)
        queue.container:SetPoint(self.anchor, UIParent, "CENTER")
    end
    queue.outside = outside

    setmetatable(queue, self)
    self.__index = self
    return queue
end

FrameQueue.padding = 1
FrameQueue.anchor = "TOP"
FrameQueue.outside = false

setmetatable(FrameQueue, { __call = FrameQueue.NewFrameQueue })
--------------------------------------------------------------------------------
ns.FrameQueue = FrameQueue
