
local ADDON_NAME, ns = ...

local Printer = ns.Printer

local PREFIX = "OAK_INTERRUPTS"

local Comms = {}

--------------------------------------------------------------------------------

local success = RegisterAddonMessagePrefix(PREFIX)
if not success then
    Printer.print("Error: Could not register for messages")
    return
end

local shared_displays = {}

local function GenRefName(disp)
    if not shared_displays[disp.name] then
        return disp.name
    else
        local count = 1
        while shared_displays[disp.name .. count] do
            count = count + 1
        end
        return disp.name .. count
    end
end

local refName = setmetatable({}, { __index = function(t, k)
                                                 local name = GenRefName(k)
                                                 t[k] = name
                                                 return name
                                             end } )

local function split(str, sep)
    local sep, fields = sep or ":", {}
    local pattern = string.format("([^%s]+)", sep)
    str:gsub(pattern, function(token) fields[#fields + 1] = token end)
    return fields
end

function Comms.HandleMessage(prefix, message, channel, sender)
    if prefix ~= PREFIX then
        return
    end
    Printer.debug(3, "Recieved Oak Interrupts message:\n%s\n%s\n%s\n%s", prefix, message, channel, sender)
    if string.match(message, "^reset .+") then
        Printer.debug(2, "Message type: reset")
        local display = shared_displays[string.sub(message, 7)]
        if display then
            ns.rotation.Reset(display)
        end
    elseif string.match(message, "^rotation ") then
        Printer.debug(2, "Message type: rotation")
        local data = string.sub(message, 10)
        local _, div_i = string.find(data, ".+:")
        local ref_disp_name = string.sub(data, 1, div_i - 1)
        Printer.debug(3, "about to check for display, ref_disp_name: %s", ref_disp_name)
        local disp = shared_displays[ref_disp_name]

        data = string.sub(data, div_i + 1)
        local order = split(data, " ")
        Printer.debug(3, "Order: ")
        for k, v in ipairs(order) do
            Printer.debug(3, "%s: %s", k, v)
        end

        if disp then -- existing shared display
            Printer.debug(2, "display already exists - %s", disp.name)
            -- clear rotation --
            for k, v in pairs(disp.watch_list) do
                if v then disp:RemoveInterrupter(k) end
            end
            -- set rotation --
            for k, v in ipairs(order) do
                disp:AddInterrupter(v)
            end
            disp:UpdateLayout()
        else -- display doesn't exist,create new display --
            Printer.debug(2, "display does not exist, building new one...")
            -- get unique name for created display
            local disp_name = not ns.displays[ref_disp_name] and ref_disp_name
                        or string.format("%s: %s", sender, ref_disp_name)
            if ns.displays[disp_name] then
                local count = 1
                while ns.displays[disp_name .. " - " .. count] do
                    count = count + 1
                end
                disp_name = disp_name .. " - " .. count
            end

            ns.display.New(disp_name, true)
            disp = ns.displays[disp_name]

            ns.display.SetRotation(disp, true)

            for k, v in ipairs(order) do
                disp:AddInterrupter(v)
            end
            disp:UpdateLayout()

            Printer.debug(3, "about to set shared_display ref_disp_name: %s", ref_disp_name)
            shared_displays[ref_disp_name] = disp
        end
    end
end

function Comms.SendReset(disp)
    local ref_disp_name = refName[disp]
    Printer.debug(2, "Sending reset for %s: %s", disp.name, ref_disp_name)
    if disp.rotation and shared_displays[ref_disp_name] then
        Printer.debug(2, "Sending rotation reset for %s to group", disp.name)
        local message = "reset " .. ref_disp_name
        SendAddonMessage(PREFIX, message, "RAID")
    end
end

function Comms.SendRotation(disp)
    if disp.rotation then
        Printer.debug(2, "Sending rotation for %s to group", disp.name)
        local names = {}
        for k, v in ipairs(disp.queue.frames) do
            tinsert(names, v.name)
        end
        local order = table.concat(names, " ")
        local ref_disp_name = refName[disp]
        local message = "rotation " .. ref_disp_name .. ":" .. order

        SendAddonMessage(PREFIX, message, "RAID")
        shared_displays[ref_disp_name] = disp
    end
end

function Comms.Cleanup(disp)
    shared_displays[refName[disp]] = nil
end
--------------------------------------------------------------------------------

ns.Comms = Comms
